from pathlib import Path
from zipfile import ZipFile

from PyQt5.QtCore import QThreadPool
from PyQt5.QtWidgets import QAction, QFileDialog

from regexport.utils.parallel import Task


class SaveBiopExtensionsActionModel:
    text = "-1. Download BIOP Extensions"

    def download_extensions(self, directory: Path):
        import requests
        if not directory.exists():
            raise FileNotFoundError(f"directory {directory} does not exist.")

        urls = [
            ('extensions', "https://github.com/BIOP/qupath-biop-extensions/releases/download/v2.0.8/biop-tools-2.0.8.jar"),
            ('extensions', "https://github.com/BIOP/qupath-biop-extensions/releases/download/v2.0.8/WSI-Dependencies.zip"),
            ('.', "https://github.com/SuperElastix/elastix/releases/download/5.0.1/elastix-5.0.1-win64.zip"),
            ('.', "https://gist.githubusercontent.com/NicoKiaru/b91f9f3f0069b765a49b5d4629a8b1c7/raw/571954a443d1e1f0597022f6c19f042aefbc0f5a/TestRegister.groovy"),
            ('.', "https://github.com/SuperElastix/elastix/releases/download/5.0.1/elastix-5.0.1-mac.zip")
        ]
        for subdir, url in urls:
            response = requests.get(url, allow_redirects=True)

            fname = directory / subdir / Path(url).name
            fname.parent.mkdir(parents=True, exist_ok=True)
            with open(fname, 'wb') as f:
                f.write(response.content)

            if fname.suffix == '.zip':
                with ZipFile(fname, 'r') as zip_ref:
                    zip_ref.extractall(fname.parent)
                fname.unlink(missing_ok=True)


class SaveBiopExtensionsAction(QAction):

    def __init__(self, model: SaveBiopExtensionsActionModel, *args, **kwargs):
        self.model = model
        super().__init__(*args, **kwargs)
        self.setText(model.text)
        self.triggered.connect(self.click)

    def click(self):
        directory = QFileDialog.getExistingDirectory(
            caption="Make a QuPath Common Files folder for your BIOP Extensions",
        )
        if not directory:
            return

        worker = Task(self.model.download_extensions, Path(directory))
        pool = QThreadPool.globalInstance()
        pool.start(worker)



import numpy as np
from traitlets import HasTraits, Instance, Unicode
from vedo import Plotter
from vedo.pyplot import histogram
from vtkmodules.qt.QVTKRenderWindowInteractor import QVTKRenderWindowInteractor

from regexport.model import AppState
from regexport.views.utils import HasWidget


class HistogramModel(HasTraits):
    data = Instance(np.ndarray, default_value=np.zeros(0))
    title = Unicode(default_value="")

    def register(self, model: AppState):
        self.model = model
        model.observe(self.update, ['cells', 'column_to_plot'])

    def update(self, change):
        model = self.model
        if model.cells is None:
            self.data = np.zeros(0)
        elif (data_column := model.cells[model.column_to_plot]).dtype.name == 'category':
            self.data = np.zeros(0)
        else:
            self.data = data_column.values

class HistogramView(HasWidget):

    def __init__(self, model: HistogramModel):

        widget = QVTKRenderWindowInteractor()
        HasWidget.__init__(self, widget=widget)
        self.plotter = Plotter(qtWidget=widget)

        self.model = model
        self.model.observe(self.render)

    def render(self, change):
        data = self.model.data
        if len(data) == 0:
            self.plotter.clear()
            return
        bin_edges = np.histogram_bin_edges(data, bins='fd')
        print('bin edges:', bin_edges)
        hist = histogram(data, bins=len(bin_edges), gap=0., title=self.model.title, )
        self.plotter.clear()
        self.plotter.show(hist, mode=12)



from PyQt5.QtWidgets import QComboBox
from traitlets import HasTraits, List as TList, Unicode, directional_link, observe

from regexport.model import AppState
from regexport.views.utils import HasWidget


class TextSelectorModel(HasTraits):
    options = TList(Unicode, default_value=["Unknown", "Unknown2"])
    selected = Unicode(default_value="Unknown")

    def register(self, model: AppState, options_attr: str, selected_attr: str):
        directional_link(
            (model, options_attr),
            (self, 'options'),
        )
        self.selected = getattr(model, selected_attr)
        directional_link(
            (self, 'selected'),
            (model, selected_attr),
        )

    @observe('options')
    def reset_selected(self, changed):
        self.selected = self.options[0]


class DropdownTextSelectorView(HasWidget):
    def __init__(self, model: TextSelectorModel):
        self.dropdown = QComboBox()
        HasWidget.__init__(self, widget=self.dropdown)

        self.model = model
        self.model.observe(self.update_dropdown_values, ['options'])
        self.model.observe(self.update_selected, ['selected'])
        self.dropdown.currentTextChanged.connect(self.select_text_from_dropdown)
        self.update_dropdown_values(None)


    def update_dropdown_values(self, changed):
        self.dropdown.currentTextChanged.disconnect()  # clearin
        self.dropdown.clear()
        for option in self.model.options:
            self.dropdown.addItem(option)
        self.dropdown.setCurrentText(self.model.selected)
        self.dropdown.currentTextChanged.connect(self.select_text_from_dropdown)

    def update_selected(self, changed):
        self.dropdown.setCurrentText(self.model.selected)

    def select_text_from_dropdown(self, text):
        print('selected', text)
        self.model.selected = text


from typing import List

from PyQt5.QtWidgets import QVBoxLayout, QWidget
from traitlets import HasTraits

from .utils import HasWidget


class SidebarModel(HasTraits):
    pass


class Sidebar(HasWidget):

    def __init__(self, model: SidebarModel, widgets: List[HasWidget] = []):
        self.model = model

        self._widget = QWidget()
        HasWidget.__init__(self, widget=self._widget)

        layout = QVBoxLayout()
        self._widget.setLayout(layout)

        self.widgets = widgets
        for widget in widgets:
            layout.addWidget(widget.widget)








from dataclasses import dataclass, field
from functools import partial
from pathlib import Path
from typing import Optional, Tuple

import numpy as np
from PyQt5.QtCore import QThreadPool
from matplotlib import pyplot as plt
from traitlets import HasTraits, Instance, directional_link
from vedo import Plotter, Mesh, Points
from vtkmodules.qt.QVTKRenderWindowInteractor import QVTKRenderWindowInteractor

from regexport.model import AppState
from regexport.utils.parallel import Task
from regexport.utils.plotting import convert_values_to_colors
from regexport.utils.profiling import warn_if_slow
from regexport.views.utils import HasWidget


@dataclass(frozen=True)
class PointCloud:
    coords: np.ndarray = field(default=np.empty((0, 3), dtype=float))
    colors: np.ndarray = field(default=np.empty((0, 3), dtype=float))
    alphas: np.ndarray = field(default=np.empty((0, 1), dtype=float))

    def __post_init__(self):
        assert self.coords.ndim == 2 and self.coords.shape[1] == 3
        assert self.colors.ndim == 2 and self.colors.shape[1] == 3
        assert self.alphas.ndim == 2 and self.alphas.shape[1] == 1


class PlotterModel(HasTraits):
    atlas_mesh = Instance(Path, allow_none=True)
    points = Instance(PointCloud, default_value=PointCloud())

    def register(self, model: AppState):
        self.model = model
        directional_link(
            (model, 'atlas'),
            (self, 'atlas_mesh'),
            lambda atlas: Path(str(atlas.structures[997]['mesh_filename'])) if atlas is not None else None
        )
        model.observe(self.link_cells_to_points, names=[
            'selected_cell_ids', 'cells', 'selected_colormap', 'column_to_plot',
        ])

    def link_cells_to_points(self, change):
        model = self.model
        if model.cells is None:
            self.points = PointCloud()
            return
        color_col = model.cells[model.column_to_plot]
        worker = Task(
            self.plot_cells,
            positions=model.cells[['X', 'Y', 'Z']].values * 1000,
            colors=color_col.cat.codes.values if color_col.dtype.name == 'category' else color_col.values,
            indices=model.selected_cell_ids if model.selected_cell_ids is not None else (),
            cmap=self.model.selected_colormap
        )
        worker.signals.finished.connect(partial(setattr, self, "points"))
        pool = QThreadPool.globalInstance()
        pool.start(worker)

    @staticmethod
    @warn_if_slow()
    def plot_cells(positions: np.ndarray, colors: np.ndarray, indices: Tuple[int], cmap: str = 'tab20c') -> PointCloud:
        return PointCloud(
            coords=positions[indices, :],
            colors=(selected_colors := convert_values_to_colors(colors, getattr(plt.cm, cmap))[indices])[:, :3],
            alphas=selected_colors[:, 3:4],
        )


class PlotterView(HasWidget):

    def __init__(self, model: PlotterModel):
        self.model = model
        self.item_points = {}
        self._atlas_mesh = None

        widget = QVTKRenderWindowInteractor()
        HasWidget.__init__(self, widget=widget)
        self.plotter = Plotter(qtWidget=widget)

        self.model.observe(self.observe_atlas_mesh, ['atlas_mesh'])
        self.model.observe(self.render, ['points'])

    @property
    def atlas_mesh(self) -> Optional[Mesh]:
        return self._atlas_mesh

    @atlas_mesh.setter
    def atlas_mesh(self, value: Optional[Mesh]):
        self._atlas_mesh = value
        self.render(None)

    @staticmethod
    def load_mesh(filename: Path) -> Mesh:
        return Mesh(
            str(filename),
            alpha=0.1,
            computeNormals=True,
            c=(1., 1., 1.)
        )

    def observe_atlas_mesh(self, change):
        print('saw atlas change')
        if self.model.atlas_mesh is None:
            self._atlas_mesh = Mesh()
        else:
            print('loading')
            worker = Task(self.load_mesh, self.model.atlas_mesh)
            worker.signals.finished.connect(partial(setattr, self, "atlas_mesh"))

            pool = QThreadPool.globalInstance()
            pool.start(worker)


    @warn_if_slow()
    def render(self, change):
        actors = [self._atlas_mesh]
        # box = self._atlas_mesh.box().wireframe().alpha(0.2).color((255, 0, 0))

        # actors.append(box)
        if len((points := self.model.points).coords) > 0:
            coords = points.coords
            colors = (np.hstack((points.colors, points.alphas)) * 255).astype(int)  # alphas needed for fast rendering.
            actors.append(Points(coords, r=3, c=colors))

        self.plotter.show(actors, at=0)
        self.plotter.addInset(self._atlas_mesh, pos=(.9, .9), size=0.1, c='w', draggable=True)
        # note: look at from vedo.applications import SlicerPlotter for inspiration


from typing import List, Union

from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QMainWindow, QWidget, QSplitter, QMenu, QAction, QToolBar

from .utils import HasWidget


class MainWindow(QMainWindow):

    def __init__(self, main_widgets: List[Union[QWidget, HasWidget]], menu_actions: List[QAction]):
        super().__init__()
        self._widgets = main_widgets  # keep reference to these widgets, so they aren't garbage collected.

        self._splitter = QSplitter(Qt.Horizontal)
        self.setCentralWidget(self._splitter)
        for widget in self._widgets:
            self._splitter.addWidget(widget.widget if hasattr(widget, 'widget') else widget)

        self._actions = menu_actions
        tool_bar = QToolBar()
        for action in menu_actions:
            tool_bar.addAction(action)
        self.addToolBar(Qt.TopToolBarArea, tool_bar)





from enum import Enum, auto
from functools import cached_property, lru_cache

import numpy as np
import pandas as pd
from bg_atlasapi import BrainGlobeAtlas
from matplotlib import pyplot as plt
from traitlets import HasTraits, Instance, observe, Tuple, List, UseEnum, Unicode

from regexport.data.filters import is_parent


class AnalysisType(Enum):
    RegionLabel = auto()
    SubCellCount = auto()


class AppState(HasTraits):
    atlas = Instance(BrainGlobeAtlas, allow_none=True)
    cells = Instance(pd.DataFrame, allow_none=True)
    selected_region_ids = Tuple(default_value=())  # should be tuple of ints
    selected_cell_ids = Instance(np.ndarray, default_value=np.array([], dtype=int))  # array of ints
    column_to_plot_options = List(Unicode(), default_value=["BrainRegion"])
    column_to_plot = Unicode(default_value="BrainRegion")
    colormap_options = List(Unicode(), default_value=[cmap for cmap in plt.colormaps() if not cmap.endswith('_r')])#['tab20c', 'viridis'])
    selected_colormap = Unicode(default_value='tab20c')

    @observe('cells')
    def _update_column_to_plot_options(self, change):
        if self.cells is None:
            return
        self.cells: pd.DataFrame
        columns = [name for name in self.cells.columns if 'Num Spots' in name]
        self.column_to_plot = "BrainRegion"
        self.column_to_plot_options = ["BrainRegion"] + columns

    @observe('column_to_plot')
    def _validate_column_to_plot(self, change):
        assert self.column_to_plot in self.column_to_plot_options

    @observe('selected_region_ids', 'cells')
    def _update_selected_cell_ids(self, change):
        if self.cells is None:
            return
        elif len(self.selected_region_ids) == 0:
            self.selected_cell_ids = self.cells.index.values
        else:
            is_parented = self.cells.groupby('BGIdx', as_index=False).BGIdx.transform(is_parent, selected_ids=self.selected_region_ids, atlas=self.atlas)
            only_parented = is_parented[is_parented.BGIdx].index.values
            self.selected_cell_ids = only_parented






from typing import Tuple

from bg_atlasapi import BrainGlobeAtlas
from pandas import Series


def is_parent(id: Series, selected_ids: Tuple[int, ...], atlas: BrainGlobeAtlas) -> bool:
    assert len(id.unique()) == 1
    id = int(id.values[0])
    if id != 0:
        if id in selected_ids:
            return True
        else:
            is_ancestor = atlas.hierarchy.is_ancestor
            return any(is_ancestor(selected_id, id) for selected_id in selected_ids)
    else:
        return False



from setuptools import setup, find_packages

long_description = """
# ABBA-QuPath Registration Exporter

A Python app and Groovy script that helps export transformed cell detections
from ABBA-registered data stored in QuPath projects.

## Installation

```
pip install ABBA-QuPath-RegistrationExporter
```

## Running

```
regexport
```
"""

setup(
    name='ABBA-QuPath-RegistrationExporter',
    version='0.2.1',
    packages=find_packages(),
    url='',
    license='MIT',
    author='Nicholas A. Del Grosso and Estelle Nassar',
    author_email='delgrosso.nick@gmail.com',
    description='A QuPath-Abba Registration Exporter Application for Mouse Sections',
    long_description=long_description,
    long_description_content_type='text/markdown',
    install_requires=[
        'numpy~=1.21.2',
        'matplotlib~=3.4.3',
        'pandas~=1.3.2',
        'qtpy~=1.11.0',
        'vtk~=9.0.3',
        'vedo~=2021.0.5',
        'pyqt5~=5.15.4',
        'treelib~=1.6.1',
        'traitlets~=5.1.0',
        'bg-atlasapi',
        'pyarrow',
    ],
    entry_points={'console_scripts':[
        "regexport=regexport.main:main"
    ]},
    package_data={
        'regexport': ['qupath_scripts/*.groovy'],
    }
)
